package shopping.cart

import akka.Done
import akka.actor.testkit.typed.scaladsl.ScalaTestWithActorTestKit
import akka.persistence.query.Offset
import akka.persistence.query.typed.EventEnvelope
import akka.projection.ProjectionId
import akka.projection.r2dbc.scaladsl.R2dbcSession
import akka.projection.scaladsl.Handler
import akka.projection.testkit.scaladsl.{
  ProjectionTestKit,
  TestProjection,
  TestSourceProvider
}
import akka.stream.scaladsl.Source
import org.scalatest.wordspec.AnyWordSpecLike
import shopping.cart.repository.ItemPopularityRepository

import java.time.Instant
import scala.concurrent.{ ExecutionContext, Future }

object ItemPopularityProjectionSpec {
  // stub out the db layer and simulate recording item count updates
  class TestItemPopularityRepository extends ItemPopularityRepository {
    var counts: Map[String, Long] = Map.empty

    override def update(
        session: R2dbcSession,
        itemId: String,
        delta: Int): Future[Long] = {
      counts = counts + (itemId -> (counts.getOrElse(itemId, 0L) + delta))
      Future.successful(1)
    }

    override def getItem(
        session: R2dbcSession,
        itemId: String): Future[Option[Long]] =
      Future.successful(counts.get(itemId))
  }
}

class ItemPopularityProjectionSpec
    extends ScalaTestWithActorTestKit
    with AnyWordSpecLike {
  import ItemPopularityProjectionSpec.TestItemPopularityRepository

  private val projectionTestKit = ProjectionTestKit(system)

  private def createEnvelope[T](
      event: T,
      seqNo: Long,
      timestamp: Long = 0L): EventEnvelope[T] =
    EventEnvelope[T](
      Offset.sequence(seqNo),
      "persistenceId",
      seqNo,
      event,
      timestamp,
      "ShoppingCart",
      1)

  private def toAsyncHandler(itemHandler: ItemPopularityProjectionHandler)(
      implicit
      ec: ExecutionContext): Handler[EventEnvelope[ShoppingCart.Event]] =
    eventEnvelope =>
      Future {
        itemHandler.process(session = null, eventEnvelope)
        Done
      }

  "The events from the Shopping Cart" should {

    "update item popularity counts by the projection" in {

      val events =
        Source(
          List[EventEnvelope[ShoppingCart.Event]](
            createEnvelope(
              ShoppingCart.ItemAdded("a7098", "bowling shoes", 1),
              0L),
            createEnvelope(
              ShoppingCart.ItemQuantityAdjusted("a7098", "bowling shoes", 2, 1),
              1L),
            createEnvelope(
              ShoppingCart
                .CheckedOut("a7098", Instant.parse("2020-01-01T12:00:00.00Z")),
              2L),
            createEnvelope(
              ShoppingCart.ItemAdded("0d12d", "akka t-shirt", 1),
              3L),
            createEnvelope(ShoppingCart.ItemAdded("0d12d", "skis", 1), 4L),
            createEnvelope(ShoppingCart.ItemRemoved("0d12d", "skis", 1), 5L),
            createEnvelope(
              ShoppingCart
                .CheckedOut("0d12d", Instant.parse("2020-01-01T12:05:00.00Z")),
              6L)))

      val repository = new TestItemPopularityRepository
      val projectionId =
        ProjectionId("item-popularity", "carts-0")
      val sourceProvider =
        TestSourceProvider[Offset, EventEnvelope[ShoppingCart.Event]](
          events,
          extractOffset = env => env.offset)
      val projection =
        TestProjection[Offset, EventEnvelope[ShoppingCart.Event]](
          projectionId,
          sourceProvider,
          () =>
            toAsyncHandler(
              new ItemPopularityProjectionHandler(
                "carts-0",
                system,
                repository))(system.executionContext))

      projectionTestKit.run(projection) {
        repository.counts shouldBe Map(
          "bowling shoes" -> 2,
          "akka t-shirt" -> 1,
          "skis" -> 0)
      }
    }
  }

}
